import random
import string

from django.contrib.auth.models import User
from django.db import models


class ApiUser(models.Model):
    def generate_token():
        return ''.join(
            random.SystemRandom().choice(
                string.ascii_letters + string.digits
            ) for _ in range(20))

    user = models.OneToOneField(
        User,
        on_delete=models.CASCADE,
    )
    token = models.SlugField(
        max_length=20,
        default=generate_token,
        editable=False,
        unique=True,
    )

    def __str__(self):
        return self.user.username



from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status
from rest_framework.response import Response

from crosswalk.authentication import AuthenticatedView
from crosswalk.exceptions import NestedAttributesError, ReservedKeyError
from crosswalk.models import Domain, Entity
from crosswalk.serializers import EntitySerializer
from crosswalk.validators import full_validation


class UpdateMatch(AuthenticatedView):

    def post(self, request, domain):
        """
        POST searches for an entity based on criteria. If only one entity is
        returned from query, it is updated. If more than one, return 403.
        """
        data = request.data.copy()

        try:
            domain = Domain.objects.get(slug=domain)
        except ObjectDoesNotExist:
            return Response(
                "Domain not found.",
                status=status.HTTP_404_NOT_FOUND
            )

        entities = Entity.objects.filter(domain=domain)
        entities = entities.filter(
            attributes__contains=data.get('block_attrs', {}))

        if entities.count() == 0:
            return Response(
                "Entity not found.",
                status=status.HTTP_404_NOT_FOUND
            )

        elif entities.count() > 1:
            return Response(
                "Found more than one entity. Be more specific?",
                status=status.HTTP_403_FORBIDDEN
            )

        entity = entities.first()
        update_attrs = data.get('update_attrs', {})

        # validate data
        try:
            full_validation(update_attrs)
        except (NestedAttributesError, ReservedKeyError):
            return Response(
                "Update data could not be validated.",
                status=status.HTTP_400_BAD_REQUEST
            )

        entity.attributes = {
            **entity.attributes,
            **update_attrs,
        }
        entity.save()

        return Response(
            {"entity": EntitySerializer(entities.first()).data},
            status=status.HTTP_200_OK
        )



from django.contrib import admin

from crosswalk.models import ApiUser


class ApiUserAdmin(admin.ModelAdmin):
    fields = ("user", "token")
    readonly_fields = ("token",)
    list_display = ("user", "token")


admin.site.register(ApiUser, ApiUserAdmin)



import uuid

from crosswalk.models import Domain
from crosswalk.validators import (validate_no_reserved_keys,
                                  validate_shallow_dict)
from django.contrib.auth.models import User
from django.contrib.postgres.fields import JSONField
from django.db import models


class Entity(models.Model):
    uuid = models.UUIDField(
        default=uuid.uuid4, editable=False, primary_key=True)

    domain = models.ForeignKey(
        Domain,
        related_name='entities',
        on_delete=models.PROTECT,
    )

    attributes = JSONField(validators=[
        validate_shallow_dict,
        validate_no_reserved_keys
    ])

    alias_for = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        related_name='aliases',
        on_delete=models.CASCADE,
        help_text="Entity in the same domain whose UUID supersedes this one.",
    )

    superseded_by = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        related_name='supersedes',
        on_delete=models.SET_NULL,
        help_text="Entity in another domain whose UUID supersedes this one.",
    )

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    created_by = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        on_delete=models.SET_NULL
    )

    @property
    def is_alias(self):
        return bool(self.alias_for)

    @property
    def is_superseded(self):
        return bool(self.superseded_by)

    def __str__(self):
        return self.attributes.get('name', self.uuid.hex)

    class Meta:
        verbose_name_plural = "entities"
        unique_together = ("domain", "attributes")



from django.urls import include, path
from rest_framework import routers

from .views import (AliasOrCreate, BestMatch, BestMatchOrCreate, BulkCreate,
                    ClientCheck, DeleteMatch, UpdateMatch)
from .viewsets import DomainViewSet, EntityDomainViewSet, EntityViewSet

router = routers.DefaultRouter()

router.register("domains", DomainViewSet, basename="domains")

entity_detail = EntityViewSet.as_view(
    {
        "get": "retrieve",
        "put": "update",
        "patch": "partial_update",
        "delete": "destroy",
    }
)
entity_domain_list = EntityDomainViewSet.as_view(
    {"get": "list", "post": "create"}
)
entity_domain_detail = EntityDomainViewSet.as_view(
    {
        "get": "retrieve",
        "put": "update",
        "patch": "partial_update",
        "delete": "destroy",
    }
)

urlpatterns = [
    path("api/", include(router.urls)),
    path(
        "api/domains/<slug:domain>/entities/",
        entity_domain_list,
        name="crosswalk-entity-domain-lists",
    ),
    path(
        "api/domains/<slug:domain>/entities/bulk-create/", BulkCreate.as_view()
    ),
    path(
        "api/domains/<slug:domain>/entities/best-match/", BestMatch.as_view()
    ),
    path(
        "api/domains/<slug:domain>/entities/best-match-or-create/",
        BestMatchOrCreate.as_view(),
    ),
    path(
        "api/domains/<slug:domain>/entities/delete-match/",
        DeleteMatch.as_view(),
    ),
    path(
        "api/domains/<slug:domain>/entities/update-match/",
        UpdateMatch.as_view(),
    ),
    path(
        "api/domains/<slug:domain>/entities/alias-or-create/",
        AliasOrCreate.as_view(),
    ),
    path(
        "api/domains/<slug:domain>/entities/<uuid:pk>/",
        entity_domain_detail,
        name="crosswalk-entity-domain-detail",
    ),
    path(
        "api/entities/<uuid:pk>/",
        entity_detail,
        name="crosswalk-entity-detail",
    ),
    path("api/client-check/", ClientCheck.as_view()),
]



from .client_check import ClientCheck  # noqa
from .bulk_create import BulkCreate  # noqa
from .best_match import BestMatch  # noqa
from .best_match_or_create import BestMatchOrCreate  # noqa
from .delete_match import DeleteMatch  # noqa
from .alias_or_create import AliasOrCreate  # noqa
from .update_match import UpdateMatch  # noqa
